@@ -8,7 +8,11 @@ use std::{
     process::Command,
 };
 
-use prettytable::{row, cell, Table, format::{LinePosition, LineSeparator}};
+use prettytable::{
+    cell,
+    format::{LinePosition, LineSeparator},
+    row, Table,
+};
 
 impl From<Int> for Expression {
     fn from(x: Int) -> Self {
@@ -34,6 +38,12 @@ impl From<String> for Expression {
     }
 }
 
+impl From<Vec<u8>> for Expression {
+    fn from(x: Vec<u8>) -> Self {
+        Self::Bytes(x)
+    }
+}
+
 impl From<bool> for Expression {
     fn from(x: bool) -> Self {
         Self::Boolean(x)
@@ -66,7 +76,7 @@ where
     }
 }
 
-#[derive(Clone)]
+#[derive(Clone, PartialEq)]
 pub enum Expression {
     Group(Box<Self>),
 
@@ -75,6 +85,8 @@ pub enum Expression {
     Integer(Int),
     // A floating point number literal
     Float(f64),
+    // A list of bytes
+    Bytes(Vec<u8>),
     // A string literal
     String(String),
     // A boolean literal
@@ -101,19 +113,39 @@ pub enum Expression {
     Macro(String, Box<Self>),
     Do(Vec<Self>),
     // A builtin function.
-    //
-    // The first argument is the name of the function, the second is the
-    // function pointer for executing the function. The third is the
-    // help string for the function.
-    Builtin(
-        String,
-        fn(Vec<Self>, &mut Environment) -> Result<Self, Error>,
-        String,
-    ),
+    Builtin(Builtin),
 
     Quote(Box<Self>),
 }
 
+#[derive(Clone)]
+pub struct Builtin {
+    /// name of the function
+    pub name: String,
+    /// function pointer for executing the function
+    pub body: fn(Vec<Expression>, &mut Environment) -> Result<Expression, Error>,
+    /// help string
+    pub help: String,
+}
+
+impl fmt::Debug for Builtin {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(f, "builtin@{}", self.name)
+    }
+}
+
+impl fmt::Display for Builtin {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(f, "builtin@{}", self.name)
+    }
+}
+
+impl PartialEq for Builtin {
+    fn eq(&self, other: &Self) -> bool {
+        self.name == other.name
+    }
+}
+
 impl fmt::Debug for Expression {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
@@ -122,6 +154,7 @@ impl fmt::Debug for Expression {
             Self::Symbol(name) => write!(f, "{}", name),
             Self::Integer(i) => write!(f, "{}", *i),
             Self::Float(n) => write!(f, "{}", *n),
+            Self::Bytes(b) => write!(f, "b{:?}", b),
             Self::String(s) => write!(f, "{:?}", s),
             Self::Boolean(b) => write!(f, "{}", if *b { "True" } else { "False" }),
             Self::List(exprs) => write!(
@@ -171,7 +204,7 @@ impl fmt::Debug for Expression {
                     .collect::<Vec<String>>()
                     .join(" ")
             ),
-            Self::Builtin(name, _, _) => write!(f, "builtin@{}", name),
+            Self::Builtin(builtin) => fmt::Debug::fmt(builtin, f),
         }
     }
 }
@@ -184,6 +217,7 @@ impl fmt::Display for Expression {
             Self::Symbol(name) => write!(f, "{}", name),
             Self::Integer(i) => write!(f, "{}", *i),
             Self::Float(n) => write!(f, "{}", *n),
+            Self::Bytes(b) => write!(f, "b{:?}", b),
             Self::String(s) => write!(f, "{}", s),
             Self::Boolean(b) => write!(f, "{}", if *b { "True" } else { "False" }),
             Self::List(exprs) => write!(
@@ -195,15 +229,6 @@ impl fmt::Display for Expression {
                     .collect::<Vec<String>>()
                     .join(", ")
             ),
-            // Self::Map(exprs) => write!(
-            //     f,
-            //     "{{{}}}",
-            //     exprs
-            //         .iter()
-            //         .map(|(k, e)| format!("{}: {:?}", k, e))
-            //         .collect::<Vec<String>>()
-            //         .join(", ")
-            // ),
             Self::Map(exprs) => {
                 let mut t = Table::new();
                 let fmt = t.get_format();
@@ -215,7 +240,7 @@ impl fmt::Display for Expression {
                 fmt.separator(LinePosition::Intern, LineSeparator::new('─', '┼', '├', '┤'));
                 fmt.separator(LinePosition::Bottom, LineSeparator::new('─', '┴', '└', '┘'));
                 for (key, val) in exprs {
-                    if let Self::Builtin(_, _, help) = &val {
+                    if let Self::Builtin(Builtin { help, .. }) = &val {
                         t.add_row(row!(key, format!("{}", val), help));
                     } else {
                         t.add_row(row!(key, format!("{}", val)));
@@ -251,26 +276,7 @@ impl fmt::Display for Expression {
                     .collect::<Vec<String>>()
                     .join(" ")
             ),
-            Self::Builtin(name, _, _) => write!(f, "builtin@{}", name),
-        }
-    }
-}
-
-impl PartialEq for Expression {
-    fn eq(&self, other: &Self) -> bool {
-        match (self, other) {
-            (Self::Symbol(a), Self::Symbol(b)) => a == b,
-            (Self::Integer(a), Self::Integer(b)) => a == b,
-            (Self::Float(a), Self::Float(b)) => a == b,
-            (Self::String(a), Self::String(b)) => a == b,
-            (Self::Boolean(a), Self::Boolean(b)) => a == b,
-            (Self::List(a), Self::List(b)) => a == b,
-            (Self::Map(a), Self::Map(b)) => a == b,
-            (Self::None, Self::None) => true,
-            (Self::Lambda(a, b, c), Self::Lambda(x, y, z)) => a == x && b == y && c == z,
-            (Self::Macro(a, b), Self::Macro(c, d)) => a == c && b == d,
-            (Self::Builtin(a, _, _), Self::Builtin(b, _, _)) => a == b,
-            _ => false,
+            Self::Builtin(builtin) => fmt::Display::fmt(builtin, f),
         }
     }
 }
@@ -281,6 +287,7 @@ impl PartialOrd for Expression {
             (Self::Integer(a), Self::Integer(b)) => a.partial_cmp(b),
             (Self::Float(a), Self::Float(b)) => a.partial_cmp(b),
             (Self::String(a), Self::String(b)) => a.partial_cmp(b),
+            (Self::Bytes(a), Self::Bytes(b)) => a.partial_cmp(b),
             (Self::List(a), Self::List(b)) => a.partial_cmp(b),
             (Self::Map(a), Self::Map(b)) => a.partial_cmp(b),
             _ => None,
@@ -294,7 +301,11 @@ impl Expression {
         body: fn(Vec<Self>, &mut Environment) -> Result<Self, Error>,
         help: impl ToString,
     ) -> Self {
-        Self::Builtin(name.to_string(), body, help.to_string())
+        Self::Builtin(Builtin {
+            name: name.to_string(),
+            body,
+            help: help.to_string(),
+        })
     }
 
     pub fn new(x: impl Into<Self>) -> Self {
@@ -306,12 +317,13 @@ impl Expression {
             Self::Integer(i) => *i != 0,
             Self::Float(f) => *f != 0.0,
             Self::String(s) => !s.is_empty(),
+            Self::Bytes(b) => !b.is_empty(),
             Self::Boolean(b) => *b,
             Self::List(exprs) => !exprs.is_empty(),
             Self::Map(exprs) => !exprs.is_empty(),
             Self::Lambda(_, _, _) => true,
             Self::Macro(_, _) => true,
-            Self::Builtin(_, _, _) => true,
+            Self::Builtin(_) => true,
             _ => false,
         }
     }
@@ -322,9 +334,10 @@ impl Expression {
             Self::None
             | Self::Integer(_)
             | Self::Float(_)
+            | Self::Bytes(_)
             | Self::String(_)
             | Self::Boolean(_)
-            | Self::Builtin(_, _, _) => vec![],
+            | Self::Builtin(_) => vec![],
 
             Self::For(_, list, body) => {
                 let mut result = vec![];
@@ -342,7 +355,7 @@ impl Expression {
             }
             Self::Map(exprs) => {
                 let mut result = vec![];
-                for (_, expr) in exprs {
+                for expr in exprs.values() {
                     result.extend(expr.get_used_symbols())
                 }
                 result
@@ -374,7 +387,7 @@ impl Expression {
         self.clone().eval_mut(env)
     }
 
-    fn eval_mut(mut self, mut env: &mut Environment) -> Result<Self, Error> {
+    fn eval_mut(mut self, env: &mut Environment) -> Result<Self, Error> {
         loop {
             match self {
                 Self::Quote(inner) => return Ok(*inner),
@@ -389,7 +402,7 @@ impl Expression {
 
                 Self::Assign(name, expr) => {
                     let x = expr.eval_mut(env)?;
-                    env.define(&name, x.clone());
+                    env.define(&name, x);
                     return Ok(Self::None);
                 }
 
@@ -422,15 +435,21 @@ impl Expression {
                     Self::Symbol(name) | Self::String(name) => {
                         let bindings = env
                             .bindings
-                            .iter()
-                            .map(|(k, v)| (k.clone(), format!("{}", v)))
+                            .clone()
+                            .into_iter()
+                            .map(|(k, v)| (k, v.to_string()))
+                            // This is to prevent environment variables from getting too large.
+                            // This causes some strange bugs on Linux: mainly it becomes
+                            // impossible to execute any program because `the argument
+                            // list is too long`.
+                            .filter(|(_, s)| s.len() <= 1024)
                             .collect::<BTreeMap<String, String>>();
 
                         match Command::new(&name)
                             .current_dir(env.get_cwd())
                             .args(
                                 args.iter()
-                                    .filter(|x| x.clone() != &Self::None)
+                                    .filter(|&x| x != &Self::None)
                                     .map(|x| Ok(format!("{}", x.clone().eval_mut(env)?)))
                                     .collect::<Result<Vec<String>, Error>>()?,
                             )
@@ -440,9 +459,10 @@ impl Expression {
                             Ok(_) => return Ok(Self::None),
                             Err(e) => {
                                 return Err(match e.kind() {
-                                    ErrorKind::NotFound => {
-                                        Error::CustomError(format!("program \"{}\" not found", name))
-                                    }
+                                    ErrorKind::NotFound => Error::CustomError(format!(
+                                        "program \"{}\" not found",
+                                        name
+                                    )),
                                     ErrorKind::PermissionDenied => Error::CustomError(format!(
                                         "permission to execute \"{}\" denied",
                                         name
@@ -454,13 +474,15 @@ impl Expression {
                     }
 
                     Self::Lambda(param, body, old_env) if args.len() == 1 => {
-                        let mut new_env = old_env.clone();
+                        let mut new_env = old_env;
+                        new_env.set_cwd(env.get_cwd());
                         new_env.define(&param, args[0].clone().eval_mut(env)?);
                         return body.eval_mut(&mut new_env);
                     }
 
                     Self::Lambda(param, body, old_env) if args.len() > 1 => {
                         let mut new_env = old_env.clone();
+                        new_env.set_cwd(env.get_cwd());
                         new_env.define(&param, args[0].clone().eval_mut(env)?);
                         self =
                             Self::Apply(Box::new(body.eval_mut(&mut new_env)?), args[1..].to_vec());
@@ -475,11 +497,11 @@ impl Expression {
                     Self::Macro(param, body) if args.len() > 1 => {
                         let x = args[0].clone().eval_mut(env)?;
                         env.define(&param, x);
-                        self = Self::Apply(Box::new(body.eval_mut(&mut env)?), args[1..].to_vec());
+                        self = Self::Apply(Box::new(body.eval_mut(env)?), args[1..].to_vec());
                     }
 
-                    Self::Builtin(_, g, _) => {
-                        return g(args.clone(), env);
+                    Self::Builtin(Builtin { body, .. }) => {
+                        return body(args.clone(), env);
                     }
 
                     _ => return Err(Error::CannotApply(*f.clone(), args.clone())),
@@ -489,6 +511,7 @@ impl Expression {
                 Self::Lambda(param, body, captured) => {
                     let mut tmp_env = captured.clone();
                     tmp_env.define(&param, Expression::None);
+                    tmp_env.set_cwd(env.get_cwd());
                     for symbol in body.get_used_symbols() {
                         if symbol != param && !captured.is_defined(&symbol) {
                             if let Some(val) = env.get(&symbol) {
@@ -496,7 +519,7 @@ impl Expression {
                             }
                         }
                     }
-                    return Ok(Self::Lambda(param.clone(), body.clone(), tmp_env));
+                    return Ok(Self::Lambda(param.clone(), body, tmp_env));
                 }
 
                 Self::List(exprs) => {
@@ -511,7 +534,7 @@ impl Expression {
                     return Ok(Self::Map(
                         exprs
                             .into_iter()
-                            .map(|(n, x)| Ok((n.clone(), x.eval_mut(env)?)))
+                            .map(|(n, x)| Ok((n, x.eval_mut(env)?)))
                             .collect::<Result<BTreeMap<String, Self>, Error>>()?,
                     ))
                 }
@@ -529,9 +552,10 @@ impl Expression {
                 | Self::Integer(_)
                 | Self::Float(_)
                 | Self::Boolean(_)
+                | Self::Bytes(_)
                 | Self::String(_)
                 | Self::Macro(_, _)
-                | Self::Builtin(_, _, _) => return Ok(self.clone()),
+                | Self::Builtin(_) => return Ok(self.clone()),
             }
         }
     }
@@ -549,6 +573,10 @@ impl Add for Expression {
             (Self::Float(m), Self::Integer(n)) => Self::Float(m + n as f64),
             (Self::Float(m), Self::Float(n)) => Self::Float(m + n),
             (Self::String(m), Self::String(n)) => Self::String(m + &n),
+            (Self::Bytes(mut a), Self::Bytes(b)) => {
+                a.extend(b);
+                Self::Bytes(a)
+            }
             (Self::List(mut a), Self::List(b)) => {
                 a.extend(b);
                 Self::List(a)
@@ -639,16 +667,15 @@ where
 
     fn index(&self, idx: T) -> &Self {
         match (self, idx.into()) {
-            (Self::Map(m), Self::Symbol(name))
-                | (Self::Map(m), Self::String(name))=> match m.get(&name) {
-                Some(val) => val,
-                None => &Self::None,
-            },
+            (Self::Map(m), Self::Symbol(name)) | (Self::Map(m), Self::String(name)) => {
+                match m.get(&name) {
+                    Some(val) => val,
+                    None => &Self::None,
+                }
+            }
 
             (Self::List(list), Self::Integer(n)) if list.len() > n as usize => &list[n as usize],
-            _ => {
-                &Self::None
-            },
+            _ => &Self::None,
         }
     }
 }
