@@ -1,6 +1,6 @@
 use nom::{
     branch::alt,
-    bytes::complete::{tag},
+    bytes::complete::tag,
     character::complete::one_of,
     combinator::{eof, map, opt},
     error::{ErrorKind, ParseError},
@@ -93,17 +93,18 @@ where
     fn or(self, other: Self) -> Self {
         match self {
             Self::InternalError => other,
-            _ => self.clone(),
+            _ => self,
         }
     }
 }
 
-const ASCII_ALPHA: &'static str = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
-const ASCII_ALPHANUMERIC: &'static str =
-    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
-const ASCII_NONZERO_DIGIT: &'static str = "123456789";
-const ASCII_DIGIT: &'static str = "0123456789";
-const ASCII_HEX_DIGIT: &'static str = "0123456789ABCDEFabcdef";
+const ASCII_ALPHA: &str = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
+const ASCII_ALPHANUMERIC: &str = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
+const ASCII_NONZERO_DIGIT: &str = "123456789";
+const ASCII_DIGIT: &str = "0123456789";
+const ASCII_HEX_DIGIT: &str = "0123456789ABCDEFabcdef";
+
+const ALLOWED_SYMBOL_PUNCTUATION: &str = "_+-.~\\/?&<>$%#^:";
 
 pub fn parse_script(input: &str, require_eof: bool) -> IResult<&str, Expression, SyntaxError> {
     let (input, _) = try_parse_ws(input)?;
@@ -164,6 +165,8 @@ fn parse_keyword(input: &str) -> IResult<&str, &str, SyntaxError> {
             tag("&&"),
             tag("||"),
             tag("//"),
+            tag(">>"),
+            tag("|"),
             tag("<"),
             tag(">"),
             tag("+"),
@@ -185,11 +188,12 @@ fn parse_symbol(input: &str) -> IResult<&str, String, SyntaxError> {
         Err(_) => {
             let old_input = input;
 
-            let (input, head) = alt((one_of(ASCII_ALPHA), one_of("_+-.~\\/?&<>$%#^=")))(input)?;
+            let (input, head) =
+                alt((one_of(ASCII_ALPHA), one_of(ALLOWED_SYMBOL_PUNCTUATION)))(input)?;
 
             let (input, tail) = many0(alt((
                 one_of(ASCII_ALPHANUMERIC),
-                one_of("_+-.~\\/?&<>$%#^="),
+                one_of(ALLOWED_SYMBOL_PUNCTUATION),
             )))(input)?;
 
             let mut result = String::from(head);
@@ -231,16 +235,13 @@ fn parse_digits(input: &str) -> IResult<&str, String, SyntaxError> {
 }
 
 fn parse_any_digits(input: &str) -> IResult<&str, String, SyntaxError> {
-    map(
-        many1(one_of(ASCII_DIGIT)),
-        |digits| {
-            let mut result = String::new();
-            for ch in digits {
-                result.push(ch)
-            }
-            result
-        },
-    )(input)
+    map(many1(one_of(ASCII_DIGIT)), |digits| {
+        let mut result = String::new();
+        for ch in digits {
+            result.push(ch)
+        }
+        result
+    })(input)
 }
 
 fn parse_integer(input: &str) -> IResult<&str, Int, SyntaxError> {
@@ -471,7 +472,7 @@ fn parse_map(input: &str) -> IResult<&str, Expression, SyntaxError> {
         tag(","),
         separated_pair(
             delimited(try_parse_ws, parse_symbol, try_parse_ws),
-            tag(":"),
+            tag("="),
             parse_expression,
         ),
     )(input)?;
@@ -599,7 +600,6 @@ fn parse_if(input: &str) -> IResult<&str, Expression, SyntaxError> {
     );
 
     Ok((input, result))
-
 }
 
 fn parse_callable(input: &str) -> IResult<&str, Expression, SyntaxError> {
@@ -607,7 +607,7 @@ fn parse_callable(input: &str) -> IResult<&str, Expression, SyntaxError> {
     let (input, arg) = match parse_symbol(input) {
         Ok((input, symbol)) => (input, symbol),
         Err(nom::Err::Failure(e)) => return Err(nom::Err::Error(e)),
-        Err(e) => return Err(e)
+        Err(e) => return Err(e),
     };
     let (input, _) = try_parse_ws(input)?;
     let (input, fn_type) = alt((tag("->"), tag("~>")))(input)?;
@@ -662,6 +662,41 @@ fn parse_apply(input: &str) -> IResult<&str, Expression, SyntaxError> {
 }
 
 fn parse_expression(input: &str) -> IResult<&str, Expression, SyntaxError> {
+    let expr_parser = parse_expression_prec_seven;
+
+    let (input, head) = expr_parser(input)?;
+
+    let (input, list) = many0(pair(
+        delimited(parse_ws, alt((tag("|"), tag(">>"))), parse_ws),
+        expr_parser,
+    ))(input)?;
+
+    if list.is_empty() {
+        return Ok((input, head));
+    }
+
+    let mut args = vec![head];
+    for (op, item) in list {
+        args.push(match op {
+            "|" => item,
+            ">>" => Expression::Apply(
+                Box::new(Expression::Symbol("redirect-out".to_string())),
+                vec![item],
+            ),
+            _ => unreachable!(),
+        })
+    }
+
+    Ok((
+        input,
+        Expression::Group(Box::new(Expression::Apply(
+            Box::new(Expression::Symbol("pipe".to_string())),
+            args,
+        ))),
+    ))
+}
+
+fn parse_expression_prec_seven(input: &str) -> IResult<&str, Expression, SyntaxError> {
     alt((
         parse_for_loop,
         parse_if,
@@ -675,7 +710,7 @@ fn parse_expression(input: &str) -> IResult<&str, Expression, SyntaxError> {
 fn parse_expression_prec_six(input: &str) -> IResult<&str, Expression, SyntaxError> {
     let expr_parser = parse_expression_prec_five;
 
-    let (input, head) = expr_parser(input)?;
+    let (input, mut head) = expr_parser(input)?;
 
     let (input, mut list) = many0(pair(
         delimited(parse_ws, alt((tag("&&"), tag("||"))), parse_ws),
@@ -686,40 +721,25 @@ fn parse_expression_prec_six(input: &str) -> IResult<&str, Expression, SyntaxErr
         return Ok((input, head));
     }
 
-    let (mut op, mut result) = list.pop().unwrap();
-    while !list.is_empty() {
-        if let Some((next_op, item)) = list.pop() {
-            let op_fun = Expression::Symbol(
-                match op {
-                    "&&" => "and",
-                    "||" => "or",
-                    _ => unreachable!(),
-                }
-                .to_string(),
-            );
+    list.reverse();
 
-            result = Expression::Group(Box::new(Expression::Apply(
-                Box::new(op_fun.clone()),
-                vec![item, result],
-            )));
+    while let Some((op, item)) = list.pop() {
+        let op_fun = Expression::Symbol(
+            match op {
+                "&&" => "and",
+                "||" => "or",
+                _ => unreachable!(),
+            }
+            .to_string(),
+        );
 
-            op = next_op;
-        }
+        head = Expression::Group(Box::new(Expression::Apply(
+            Box::new(op_fun.clone()),
+            vec![head, item],
+        )));
     }
 
-    let op_fun = Expression::Symbol(
-        match op {
-            "&&" => "and",
-            "||" => "or",
-            _ => unreachable!(),
-        }
-        .to_string(),
-    );
-
-    Ok((
-        input,
-        Expression::Apply(Box::new(op_fun.clone()), vec![head, result]),
-    ))
+    Ok((input, head))
 }
 
 fn parse_range(input: &str) -> IResult<&str, Expression, SyntaxError> {
@@ -755,7 +775,7 @@ fn parse_expression_prec_five(input: &str) -> IResult<&str, Expression, SyntaxEr
         return Ok(result);
     }
 
-    let (input, head) = expr_parser(input)?;
+    let (input, mut head) = expr_parser(input)?;
 
     let (input, mut list) = many0(pair(
         delimited(
@@ -781,54 +801,35 @@ fn parse_expression_prec_five(input: &str) -> IResult<&str, Expression, SyntaxEr
         return Ok((input, head));
     }
 
-    let (mut op, mut result) = list.pop().unwrap();
-    while !list.is_empty() {
-        if let Some((next_op, item)) = list.pop() {
-            let op_fun = Expression::Symbol(
-                match op {
-                    "==" => "eq",
-                    "!=" => "neq",
-                    ">=" => "gte",
-                    "<=" => "lte",
-                    ">" => "gt",
-                    "<" => "lt",
-                    _ => unreachable!(),
-                }
-                .to_string(),
-            );
+    list.reverse();
 
-            result = Expression::Group(Box::new(Expression::Apply(
-                Box::new(op_fun.clone()),
-                vec![item, result],
-            )));
+    while let Some((op, item)) = list.pop() {
+        let op_fun = Expression::Symbol(
+            match op {
+                "==" => "eq",
+                "!=" => "neq",
+                ">=" => "gte",
+                "<=" => "lte",
+                ">" => "gt",
+                "<" => "lt",
+                _ => unreachable!(),
+            }
+            .to_string(),
+        );
 
-            op = next_op;
-        }
+        head = Expression::Group(Box::new(Expression::Apply(
+            Box::new(op_fun.clone()),
+            vec![head, item],
+        )));
     }
 
-    let op_fun = Expression::Symbol(
-        match op {
-            "==" => "eq",
-            "!=" => "neq",
-            ">=" => "gte",
-            "<=" => "lte",
-            ">" => "gt",
-            "<" => "lt",
-            _ => unreachable!(),
-        }
-        .to_string(),
-    );
-
-    Ok((
-        input,
-        Expression::Apply(Box::new(op_fun.clone()), vec![head, result]),
-    ))
+    Ok((input, head))
 }
 
 fn parse_expression_prec_four(input: &str) -> IResult<&str, Expression, SyntaxError> {
     let expr_parser = parse_expression_prec_three;
 
-    let (input, head) = expr_parser(input)?;
+    let (input, mut head) = expr_parser(input)?;
 
     let (input, mut list) = many0(pair(
         delimited(parse_ws, alt((tag("+"), tag("-"))), parse_ws),
@@ -839,49 +840,31 @@ fn parse_expression_prec_four(input: &str) -> IResult<&str, Expression, SyntaxEr
         return Ok((input, head));
     }
 
-    let (mut op, mut result) = list.pop().unwrap();
-    while !list.is_empty() {
-        if let Some((next_op, item)) = list.pop() {
-            let op_fun = Expression::Symbol(
-                match op {
-                    "+" => "add",
-                    "-" => "sub",
-                    _ => unreachable!(),
-                }
-                .to_string(),
-            );
+    list.reverse();
 
-            result = Expression::Group(Box::new(Expression::Apply(
-                Box::new(op_fun.clone()),
-                vec![item, result],
-            )));
+    while let Some((op, item)) = list.pop() {
+        let op_fun = Expression::Symbol(
+            match op {
+                "+" => "add",
+                "-" => "sub",
+                _ => unreachable!(),
+            }
+            .to_string(),
+        );
 
-            op = next_op;
-        }
+        head = Expression::Group(Box::new(Expression::Apply(
+            Box::new(op_fun.clone()),
+            vec![head, item],
+        )));
     }
 
-    let op_fun = Expression::Symbol(
-        match op {
-            "+" => "add",
-            "-" => "sub",
-            _ => unreachable!(),
-        }
-        .to_string(),
-    );
-
-    Ok((
-        input,
-        Expression::Group(Box::new(Expression::Apply(
-            Box::new(op_fun.clone()),
-            vec![head, result],
-        ))),
-    ))
+    Ok((input, head))
 }
 
 fn parse_expression_prec_three(input: &str) -> IResult<&str, Expression, SyntaxError> {
     let expr_parser = parse_expression_prec_two;
 
-    let (input, head) = expr_parser(input)?;
+    let (input, mut head) = expr_parser(input)?;
 
     let (input, mut list) = many0(pair(
         delimited(parse_ws, alt((tag("*"), tag("//"), tag("%"))), parse_ws),
@@ -892,52 +875,31 @@ fn parse_expression_prec_three(input: &str) -> IResult<&str, Expression, SyntaxE
         return Ok((input, head));
     }
 
-    let (mut op, mut result) = list.pop().unwrap();
-    while !list.is_empty() {
-        if let Some((next_op, item)) = list.pop() {
-            let op_fun = Expression::Symbol(
-                match op {
-                    "*" => "mul",
-                    "//" => "div",
-                    "%" => "rem",
-                    _ => unreachable!(),
-                }
-                .to_string(),
-            );
+    list.reverse();
 
-            result = Expression::Group(Box::new(Expression::Apply(
-                Box::new(op_fun.clone()),
-                vec![item, result],
-            )));
+    while let Some((op, item)) = list.pop() {
+        let op_fun = Expression::Symbol(
+            match op {
+                "*" => "mul",
+                "//" => "div",
+                "%" => "rem",
+                _ => unreachable!(),
+            }
+            .to_string(),
+        );
 
-            op = next_op;
-        }
+        head = Expression::Group(Box::new(Expression::Apply(
+            Box::new(op_fun.clone()),
+            vec![head, item],
+        )));
     }
 
-    let op_fun = Expression::Symbol(
-        match op {
-            "*" => "mul",
-            "//" => "div",
-            "%" => "rem",
-            _ => unreachable!(),
-        }
-        .to_string(),
-    );
-
-    Ok((
-        input,
-        Expression::Group(Box::new(Expression::Apply(
-            Box::new(op_fun.clone()),
-            vec![head, result],
-        ))),
-    ))
+    Ok((input, head))
 }
 
 fn parse_expression_prec_two(input: &str) -> IResult<&str, Expression, SyntaxError> {
     let (input, head) = parse_expression_prec_one(input)?;
-    let (input, args) = many0(preceded(
-        tag("@"), parse_expression_prec_one,
-    ))(input)?;
+    let (input, args) = many0(preceded(tag("@"), parse_expression_prec_one))(input)?;
 
     if args.is_empty() {
         return Ok((input, head));
@@ -960,11 +922,11 @@ fn parse_expression_prec_one(input: &str) -> IResult<&str, Expression, SyntaxErr
         parse_map,
         parse_block,
         parse_list,
-        map(parse_boolean, |b| Expression::Boolean(b)),
+        map(parse_boolean, Expression::Boolean),
         map(parse_none, |_| Expression::None),
-        map(parse_float, |n| Expression::Float(n)),
-        map(parse_integer, |n| Expression::Integer(n)),
-        map(parse_string, |x| Expression::String(x)),
-        map(parse_symbol, |x| Expression::Symbol(x)),
+        map(parse_float, Expression::Float),
+        map(parse_integer, Expression::Integer),
+        map(parse_string, Expression::String),
+        map(parse_symbol, Expression::Symbol),
     ))(input)
 }
